"""Wrap models using instancelib and instancelib-onnx."""

from pathlib import Path
from typing import Dict, Optional, Union

from instancelib.environment.base import Environment
from instancelib.instances.base import InstanceProvider
from instancelib.machinelearning import AbstractClassifier, SkLearnDataClassifier
from instancelib.typehints import LT
from sklearn.base import is_classifier
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import Pipeline
from sklearn.utils.estimator_checks import check_estimator
from sklearn.utils.validation import check_is_fitted

from ..data import train_test_split
from ..utils import get_file_type, info, package_available


def sklearn_model(model) -> bool:
    """Check is a model is an scikit-learn model.

    Args:
        model: Model to check.

    Returns:
        bool: Model is a scikit-learn model (True) or not (False).
    """
    if isinstance(model, Pipeline):
        return True
    try:
        check_estimator(model)
        return True
    except:
        return False


def sklearn_fitted(model) -> bool:
    """Check if a scikit-learn model is fitted.

    Args:
        model: Model to check.

    Returns:
        bool: Model is fitted (True) or not (False).
    """
    try:
        if isinstance(model, Pipeline):
            if all(sklearn_fitted(part) for _, part in model.steps):
                return True
        check_is_fitted(model)
        return True
    except NotFittedError:
        return False


def import_model(model,
                 environment: Optional[Environment] = None,
                 train: Union[int, float, str, InstanceProvider] = 'train',
                 label_map: Optional[Dict[LT, LT]] = None) -> AbstractClassifier:
    """Import a model from file or from a Python object.

    Examples:
        Make a scikit-learn text classifier and train it on SST2

        >>> from genbase import import_data, import_model
        >>> from datasets import load_dataset
        >>> ds = import_data(load_dataset('glue', 'sst2'), data_cols='sentence', label_cols='label')
        >>> from sklearn.pipeline import Pipeline
        >>> from sklearn.naive_bayes import MultinomialNB
        >>> from sklearn.feature_extraction.text import TfidfVectorizer
        >>> pipeline = Pipeline([('tfidf', TfidfVectorizer()),
        ...                      ('clf', MultinomialNB())])
        >>> import_model(pipeline, ds, train='train')

        Load a pretrained ONNX model downloaded from 
            https://github.com/mpbron/instancelib-onnx/blob/main/example_models/data-model.onnx

        >>> from genbase import import_model
        >>> import_model('data-model.onnx', label_map={0: 'Bedrijfsnieuws', 1: 'Games', 2: 'Smartphones'})

    Args:
        model: Model or path to model to import.
        environment (Optional[Environment], optional): Environment corresponding to model (with dataset and ground-truth
            labels), used for importing models and/or training them.
        train (Union[int, float, str, InstanceProvider], optional): Train split size, name in environment or provider. 
            Defaults to 'train'.
        label_map (Optional[Dict[LT, LT]], optional): Conversion of label IDs to named labels. Defaults to None.

    Raises:
        ImportError: Unable to import model or file.
        NotImplementedError: Type of model is not yet supported.

    Returns:
        AbstractClassifier: Instancelib wrapped model.
    """
    if label_map is None and environment is not None:
        label_map = list(environment.labels.labelset)
    if isinstance(label_map, dict):
        label_map = {str(k): v for k, v in label_map.items()}

    if isinstance(model, str):
        if not Path(model).exists():
            raise ImportError(f'Unable to locate file "{model}"')
        file_type = get_file_type(model)
        if file_type == '.pkl':
            import pickle  # nosec
            info('Unpickling model (warning: be sure you trust a source before unpickling a model!)')
            with open(model, 'rb') as f:
                model = pickle.load(f)  # nosec
        elif file_type == '.onnx':
            if not package_available('ilonnx'):
                raise ImportError('To import ONNX files install `instancelib-onnx`!')
            import ilonnx
            if label_map is None:
                info('Improve the informativeness of your predictions by providing the label_map')
            return ilonnx.build_data_model(model, classes=label_map)
    elif isinstance(model, AbstractClassifier):
        return model

    _no_train_msg = ''

    if environment is not None:
        if isinstance(train, (float, int)):
            _no_train_msg = 'Splitting dataset into train/test set...'
            environment = train_test_split(environment, train_size=train)
            train = environment['train']
        elif train in environment.keys():
            _no_train_msg = f'Using named_provider "{train}" as the training set...'
            train = environment[train]
        elif not isinstance(train, InstanceProvider):
            _no_train_msg = 'No training set provided, defaulting to all data as training set...'
            train = environment.dataset

    if sklearn_model(model):
        if not sklearn_fitted(model):
            if environment is None:
                raise ImportError('Untrained scikit-learn models require an environment to import!')   
            info('Model is not fitted yet, fitting model!')
            if _no_train_msg:
                info(_no_train_msg)
            if is_classifier(model):
                model = SkLearnDataClassifier.build(model, environment)
                model.fit_provider(train, environment.labels)
                return model
            else:
                NotImplementedError('Only classifiers are currently supported!')
        else:
            if is_classifier(model):
                classes = label_map if environment is None else environment
                return SkLearnDataClassifier.build_from_model(model, classes=classes)
            else:
                raise NotImplementedError('Only classifiers are currently supported!')
    elif isinstance(model, AbstractClassifier):
        return model
    elif 'torch' in str(type(model)):
        raise ImportError('Convert your PyTorch model with ONNX (https://pytorch.org/docs/stable/onnx.html)' +
                          ' before importing it with instancelib-onnx.')
    elif 'keras' in str(type(model)):
        raise ImportError('Convert your Keras model with ONNX (https://github.com/onnx/keras-onnx)' +
                          ' before importing it with instancelib-onnx.')
    elif 'tensorflow' in str(type(model)):
        raise ImportError('Convert your Tensorflow model with ONNX (https://github.com/onnx/tensorflow-onnx)' +
                          ' before importing it with instancelib-onnx.')        
    raise NotImplementedError(f'Unknown type of model "{model}!"')



"""Default classes for all to inherit from."""

import builtins
from pathlib import Path
from typing import List, Optional

import srsly

from genbase._version import __version__
from genbase.data import import_data, rename_labels, train_test_split
from genbase.decorator import add_callargs
from genbase.internationalization import LOCALE_MAP, get_locale, set_locale, translate_list, translate_string
from genbase.mixin import CaseMixin, SeedMixin
from genbase.model import import_model
from genbase.ui import Render, is_colab, is_interactive
from genbase.utils import recursive_to_dict, silence_tqdm


class Readable:
    """Ensure that a class has a readable representation."""

    def __repr__(self):
        public_vars = ', '.join([f'{k}={v}' for k, v in vars(self).items() if not k.startswith('_')])
        return f'{self.__class__.__name__}({public_vars})'


class Configurable:
    """Add working with configs (configuration dictionaries) to a class."""

    @classmethod
    def from_config(cls, config: dict, **kwargs) -> 'Configurable':
        config = {**config, **kwargs}
        _ = config.pop('__class__', None)
        return cls(**config)

    @classmethod
    def read_json(cls, path: str, **read_args) -> 'Configurable':
        """Read config from JSON file (GZIP JSON, JSONL or JSON).

        Args:
            path (str): File path.
            **read_args: Optional arguments passed to `srsly.read_json()`/`srsly.read_jsonl()`/`srsly.read_gzip_json`.
        """
        read_fn = srsly.read_json
        if path.endswith('.json.gz'):
            read_fn = srsly.read_gzip_json
        elif path.endswith('.jsonl'):
            read_fn = srsly.read_jsonl
        return cls.from_config(read_fn(path, **read_args))

    @classmethod
    def from_json(cls, json_or_path: str, **read_args) -> 'Configurable':
        """Get config from JSON string or filepath.

        Args:
            json_or_path (str): File path or JSON string.
            **read_args: Optional arguments passed to `srsly.read_json()`/`srsly.read_jsonl()`/`srsly.read_gzip_json`.
        """
        if Path.is_file(json_or_path):
            return cls.read_json(json_or_path, **read_args)
        return cls.from_config(srsly.json_loads(json_or_path))

    @classmethod
    def read_yaml(cls, path: str) -> 'Configurable':
        """Read config from YAML file.

        Args:
            path (str): File path.
        """
        return cls.from_config(srsly.read_yaml(path))

    @classmethod
    def from_yaml(cls, yaml_or_path: str) -> 'Configurable':
        """Get config from YAML string or filepath.

        Args:
            yaml_or_path (str): File path or YAML string.
        """
        if Path.is_file(yaml_or_path):
            return cls.read_yaml(yaml_or_path)
        return cls.from_config(srsly.yaml_loads(yaml_or_path))

    def to_config(self, exclude: List[str]) -> dict:
        """Convert class information into config (configuration dictionary).

        Args:
            exclude (List[str]): Names of variables to exclude.

        Returns:
            dict: [description]
        """
        return dict(recursive_to_dict(self, exclude=exclude))

    def to_json(self, indent: int = 2) -> str:
        """Convert config to JSON-formatted string.

        Args:
            indent (int, optional): Number of spaces to indent JSON. Defaults to 2.

        Returns:
            str: Config formatted as JSON.
        """
        return srsly.json_dumps(self.to_config(), indent=indent)

    def to_yaml(self, **write_args) -> str:
        """Convert config to YAML-formatted string.

        Args:
            **write_args: Optional arguments passed to `srsly.yaml_dumps()`

        Returns:
            str: Config formatted as YAML.
        """
        return srsly.yaml_dumps(self.to_config(), **write_args)

    def write_json(self, path: str, indent: int = 2) -> None:
        """Write class config to JSON.

        Args:
            path (str): Path to save to. If ends in `.json.gz` saves as GZIP JSON, `.jsonl` as JSONL or JSON by default.
            indent (int, optional): Number of spaces to indent JSON. Defaults to 2.
        """
        write_fn = srsly.write_json
        if path.endswith('.json.gz'):
            write_fn = srsly.write_gzip_json
        elif path.endswith('.jsonl'):
            write_fn = srsly.write_jsonl
        write_fn(path, self.to_config(), indent=indent)

    def write_yaml(self, path: str, **write_args) -> None:
        """Write class config to YAML.

        Args:
            path (str): Path to save to.
            **write_args: Optional arguments passed to `srsly.write_yaml()`
        """
        srsly.write_yaml(path, self.to_config(), **write_args)


class MetaInfo(Configurable):
    def __init__(self,
                 type: str,
                 subtype: Optional[str] = None,
                 fn_name: Optional[str] = None,
                 callargs: Optional[dict] = None,
                 renderargs: Optional[dict] = None,
                 renderer = Render,
                 **kwargs):
        """Meta information class.

        Args:
            type (str): Type description.
            subtype (Optional[str], optional): Subtype description. Defaults to None.
            callargs (Optional[dict], optional): Arguments used when the function was called. Defaults to None.
            renderargs (Optional[dict], optional): Custom arguments passed to renderer. Defaults to None.
            renderer
            **kwargs: Optional meta descriptors.
        """
        self._type = type
        self._subtype = subtype
        self._callargs = callargs
        self._renderargs = renderargs
        self._dict = {'type': type}
        if self._subtype is not None:
            self._dict['subtype'] = self._subtype
        if fn_name is not None:
            self._callargs['__name__'] = fn_name
        if self._callargs is not None:
            self._dict['callargs'] = self._callargs
        if self._renderargs is not None:
            self._dict['renderargs'] = self._renderargs
        self._dict = dict(self._dict, **kwargs)
        self._renderer = renderer if isinstance(renderer, builtins.type) else renderer.__class__

    @property
    def type(self):
        return self._type

    @property
    def subtype(self):
        return self._subtype

    @property
    def callargs(self):
        return self._callargs

    @property
    def meta(self):
        return self._dict

    @property
    def renderargs(self):
        return self._renderargs if self._renderargs is not None else {}

    @property
    def html(self):
        if is_colab():
            return self.raw_html
        if 'add_plotly' not in self.renderargs:
            self.renderargs['add_plotly'] = not is_interactive()
        return self._renderer(self.to_config()).as_html(**self.renderargs)

    @property
    def raw_html(self):
        renderargs = self.renderargs
        renderargs['add_plotly'] = True
        return self._renderer(self.to_config()).as_html(**renderargs)

    def to_config(self):
        if hasattr(self, 'content'):
            _content = self.content() if callable(self.content) else self.content
            content = dict(recursive_to_dict(_content, include_class=False))
        else:
            content = super().to_config(exclude=['_type', '_subtype', '_dict', '_callargs'])

        return {'META': self.meta, 'CONTENT': content}

    def _repr_html_(self) -> str:
        return self.html if is_interactive() else repr(self)



"""SVG icons."""

# Source: https://refokus-copy-to-clipboard.webflow.io/
CLONE = """
<svg class="icon icon-clone" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="#585B6C" stroke-width="2" xmlns="http://www.w3.org/2000/svg">
<path d="M17 6V5C17 3.34315 15.6569 2 14 2H6C4.34315 2 3 3.34315 3 5V15C3 16.6569 4.34315 18 6 18H7"/>
<rect x="7" y="6" width="14" height="16" rx="3"/>
</svg>
"""
CHECK = """
<svg class="icon icon-check" width="16" height="16" viewBox="0 0 16 16" fill="none" stroke="#28A745" strokeWidth="2" strokeLinecap="round" strokeLinejoin="round" xmlns="http://www.w3.org/2000/svg">
<path d="M13.25 4.75L6 12L2.75 8.75"/>
</svg>
"""



"""Extensible user interfaces (UIs) for `genbase` dependencies."""

from typing import Sequence, Union

import matplotlib.cm
import matplotlib.colors

from genbase.ui.notebook import Render, format_instances, is_colab, is_interactive
from genbase.ui.plot import matplotlib_available


def get_color(value: Union[float, Sequence[float]],
              min_value: float = -1.0,
              max_value: float = 1.0,
              colorscale: Union[list, str] = 'RdYlGn',
              format: str = 'hex') -> Union[str, Sequence[str]]:
    """Get color from a `matplotlib` colorscale.

    Args:
        value (Union[float, Sequence[float]]): Value(s) to convert. Will be clamped to `[min_value, max_value]`.
        min_value (float, optional): Minimum scale value. Defaults to -1.0.
        max_value (float, optional): Maximum scale value. Defaults to 1.0.
        colorscale (Union[list, str], optional): `matplotlib` scale definition or colorscale name. Defaults to 'RdYlGn'.
        format (str, optional): Return format 'hex'/'rgb'. Defaults to 'hex'.

    Raises:
        ValueError: `min_value` is not smaller than `max_value`, unknown format ().
        ImportError: Cannot import `matplotlib`.

    Returns:
        Union[str, Sequence[str]]: Named color(s) in selected format.
    """
    if min_value >= max_value:
        raise ValueError(f'min_value should be smaller than max_value, but is {min_value} ({max_value=})')

    is_float = False

    if isinstance(value, float):
        is_float = True
        value = [value]
    if not isinstance(value, list):
        value = list(value)

    if not matplotlib_available():
        raise ImportError('Currently requires `matplotlib` to be installed!')

    # Clamp value
    value = [min(max(v, min_value), max_value) for v in value]

    if isinstance(colorscale, str):
        cmap = matplotlib.cm.get_cmap(colorscale)
    else:
        cmap = matplotlib.colors.LinearSegmentedColormap.from_list('', colorscale)

    res = [cmap(int((v - min_value) / (max_value - min_value) * cmap.N)) for v in value]
    if format == 'hex':
        res = [matplotlib.colors.rgb2hex(v) for v in res]
    return res[0] if is_float else res



"""Support for i18n internationalization."""

import os
from typing import List

import i18n
from lazy_load import lazy_func

LOCALE_MAP = {'br': 'pt_BR',
              'cs': 'cs_CZ',
              'da': 'da_DK',
              'el': 'el_GR',
              'ph': 'fil_PH',
              'fr': 'fr_FR',
              'ga': 'ga_IE',
              'hi': 'hi_IN',
              'hr': 'hr_HR',
              'hu': 'hu_HU',
              'id': 'id_ID',
              'it': 'it_IT',
              'jp': 'ja_JP',
              'ka': 'ka_GE',
              'lt': 'lt_LT',
              'lv': 'lv_LV',
              'nl': 'nl_NL',
              'no': 'no_NO',
              'pl': 'pl_PL',
              'pt': 'pt_PT',
              'ro': 'ro_RO',
              'ru': 'ru_RU',
              'sk': 'sk_SK',
              'tr': 'tr_TR',
              'uk': 'uk_UA'}


FOLDER = os.path.abspath(os.path.dirname(os.path.abspath(__file__)))
if not os.path.isdir(f'{FOLDER}{os.path.sep}locale'):
    FOLDER = os.path.join(FOLDER, 'genbase')

i18n.load_path.append(os.path.join(FOLDER, 'locale'))
i18n.set('filename_format', '{locale}.{format}')
i18n.set('file_format', 'json')
i18n.set('locale', 'nl')
i18n.set('fallback', 'en')
i18n.set('skip_locale_root_data', True)
i18n.resource_loader.init_json_loader()


@lazy_func
def translate_string(id: str) -> str:
    """Get a string based on `locale`, as defined in the './locale' folder.

    Args:
        id (str): Identifier of string in `lang.{locale}.yml` file.

    Returns:
        str: String corresponding to locale.
    """
    return i18n.t(f'{id}')


@lazy_func
def translate_list(id: str, sep: str = ';') -> List[str]:
    """Get a list based on `locale`, as defined in the './locale' folder.

    Args:
        id (str): Identifier of list in `lang.{locale}.yml` file.
        sep (str, optional): Separator to split elements of list. Defaults to ';'.

    Returns:
        List[str]: List corresponding to locale.
    """
    return i18n.t(f'{id}').split(sep)


def set_locale(locale: str) -> None:
    """Set current locale (choose from `en`, `nl`).

    Args:
        locale (str): Locale to change to.
    """
    return i18n.set('locale', locale)


@lazy_func
def get_locale() -> str:
    """Get current locale.

    Returns:
        str: Current locale.
    """
    return i18n.get('locale')



import pytest

from genbase.internationalization import get_locale, set_locale, translate_list, translate_string

locale = ['nl', 'en']
ids = [i for i in range(len(locale))]

def test_swap_locale():
    set_locale(locale[0])
    assert get_locale() == locale[0]
    set_locale(locale[1])
    assert get_locale() == locale[1]

@pytest.mark.parametrize('id', ids)
def test_translate_string(id):
    set_locale(locale[id])
    assert isinstance(translate_string('str1'), str)

@pytest.mark.parametrize('id', ids)
def test_translate_string_length(id):
    set_locale(locale[id])
    assert len(translate_string('str1')) > 0

@pytest.mark.parametrize('id', ids)
def test_translate_list(id):
    set_locale(locale[id])
    assert isinstance(translate_list('list1'), list)

@pytest.mark.parametrize('id', ids)
def test_translate_list_length(id):
    set_locale(locale[id])
    assert len(translate_list('list1')) > 0



"""Base support for decorators."""

import inspect
from functools import wraps

from genbase.utils import recursive_to_dict


def add_callargs(function):
    """Decorator that passes `__callargs__`  to a function if available.

    Useful in conjunction with `genbase.MetaInfo`.

    Args:
        function: Function to wrap
    """  # noqa: D401
    @wraps(function)
    def inner(*args, **kwargs):
        ba = inspect.signature(function).bind(*args, **kwargs)
        ba.apply_defaults()

        kw = next((k for k, v in ba.signature.parameters.items()
                   if k == '__callargs__' or v.kind == inspect._ParameterKind.VAR_KEYWORD),
                  None)

        # Do not decorate the function if we are unable to pass __callargs__ as an argument
        if kw is None:
            return function(*ba.args, **ba.kwargs)

        # Construct __callargs__ (including introspection of the self argument)
        callargs = {'__name__': function.__name__, **dict(recursive_to_dict(ba.arguments))}
        if hasattr(function, '__self__') and 'self' not in callargs.keys():
            callargs['self'] = function.__self__
        if 'self' in callargs.keys():
            self = callargs.pop('self')
            callargs['self'] = self.to_config() if hasattr(self, 'to_config') and hasattr(self, '_dict') \
                else dict(recursive_to_dict(self))
            if '__name__' not in callargs['self'] and hasattr(self, '__class__') or hasattr(self, '__name__'):
                callargs['self']['__name__'] = self.__class__.__name__ if hasattr(self, '__class__') \
                    else self.__name__

        callargs.pop('__class__', None)
        return function(*ba.args, __callargs__=callargs, **ba.kwargs)
    return inner



"""Jupyter notebook rendering interface."""

import copy
import traceback
import uuid
from typing import Callable, Iterable, List, Optional, Union

import srsly
from IPython import get_ipython

from .plot import plotly_available
from .svg import CLONE as CLONE_SVG

PACKAGE_LINK = 'https://git.science.uu.nl/m.j.robeer/genbase/'
MAIN_COLOR = '#000000'
CUSTOM_CSS = """
#--var(ui_id),
#--var(ui_id) + footer,
#--var(ui_id) + svg + footer {
    -webkit-text-size-adjust: 100%;
    -webkit-tap-highlight-color: rgba(0, 0, 0, 0);
    -moz-osx-font-smoothing: grayscale;
    -webkit-font-smoothing: antialiased;
    margin: 1em;
    font-family: sans-serif;
    font-size: 1.2rem;
    line-height: 1.2;
}

#--var(ui_id) h1,
#--var(ui_id) h2,
#--var(ui_id) h3,
#--var(ui_id) h4,
#--var(ui_id) h5,
#--var(ui_id) h6 {
    color: #0d0d0d;
    line-height: 1.2;
}

#--var(ui_id) + footer a,
#--var(ui_id) + footer a:visited,
#--var(ui_id) + svg + footer a,
#--var(ui_id) + svg + footer a:visited,
#--var(ui_id) a,
#--var(ui_id) a:visited {
    background-color: transparent;
    color: --var(ui_color);
    line-height: 1.6;
}

#--var(ui_id) + footer a:hover,
#--var(ui_id) + footer a:active,
#--var(ui_id) + svg + footer a:hover,
#--var(ui_id) + svg + footer a:active,
#--var(ui_id) a:hover,
#--var(ui_id) a:active {
    border-bottom: none;
    outline: 0;
}

#--var(ui_id) + footer a:focus,
#--var(ui_id) + svg + footer a:focus,
#--var(ui_id) a:focus {
    border-bottom: none;
    outline: thin dotted;
}

#--var(ui_id) + footer a img,
#--var(ui_id) + svg + footer a img,
#--var(ui_id) a img {
    border: 0;
}

#--var(ui_id) + footer,
#--var(ui_id) + svg + footer {
    text-align: right;
    margin: 0 1rem;
    font-size: 1rem;
    color: #999;
}

#--var(ui_id) .ui-container {
    padding: 0.2rem;
}

#--var(ui_id) .ui-block {
    display: flex;
    align-items: center;
    justify-content: center;
}

#--var(tabs_id) {
    display: flex;
    flex-wrap: wrap;
    width: 100%;
    box-shadow: 0 8px 8px rgba(0, 0, 0, 0.4);
}

#--var(tabs_id) label {
    width: 100%;
    display: flex;
    align-items: center;
    justify-content: center;
    padding: 1rem 2rem;
    margin-right: 0.0625rem;
    cursor: pointer;
    background-color: --var(ui_color);
    color: #fff;
    font-size: 1.2rem;
    font-weight: 700;
    transition: background-color ease 0.3s;
}

#--var(tabs_id) .tab {
    flex-grow: 1;
    width: 100%;
    height: 100%;
    display: none;
    padding: 1rem 2rem;
    color: #000;
    background-color: #fff;
}

#--var(tabs_id) .tab > *:not(:last-child) {
    margin-bottom: 0.8rem;
}

#--var(tabs_id) [type=radio] {
    display: none;
}

#--var(tabs_id) kbd {
    color: #efefef;
    background-color: #151515;
}

#--var(tabs_id) [type=radio]:checked + label {
    background-color: #fff;
    color: --var(ui_color);
    border-top: 4px solid --var(ui_color);
}

#--var(tabs_id) [type=radio]:checked + label + .tab {
    display: block;
}

#--var(tabs_id) .code > pre {
    color: #111;
    font-family: Consolas, monospace;
    background-color: #eee !important;
    box-sizing: content-box;
    padding: 0.5rem 0.3rem !important;
    max-height: 30rem;
    overflow-x: hidden;
    overflow-y: scroll;
    box-shadow: inset 0 4px 4px rgba(0, 0, 0, 0.15);
}

#--var(tabs_id) .code section {
    position: relative;
}

#--var(tabs_id) .code section:not(:first-child) {
    margin-top: 2rem;
}

#--var(tabs_id) .code h3 {
    font-size: 18px;
    display: inline;
}

#--var(tabs_id) .code .pre-buttons {
    position: inherit;
    float: right;
}

#--var(tabs_id) .code .pre-buttons > a {
    position: relative;
    all: unset;
    height: 20px;
    width: 20px;
    background: none;
    font: inherit;
    outline: inherit;
    display: block;
}

#--var(tabs_id) .code .pre-buttons > a:hover {
    color: --var(ui_color);
}

#--var(tabs_id) .code a > svg {
    position: absolute;
    top: 0px;
    left: 0px;
    height: 18px;
}

#--var(tabs_id) .code a.copy:before {
    content: 'Copied to clipboard!';
    opacity: 0;
    position: fixed;
    bottom: 8px;
    right: 8px;
    padding: 1rem;
    color: #fff;
    background-color: #28a745;
    box-shadow: 0 4px 4px rgba(0, 0, 0, 0.4);
    transition: all 0.3s ease-in-out 2s;
}

#--var(tabs_id) .code a:active::before {
    opacity: 1;
    transition: 0s;
}

#--var(tabs_id) .code a:hover > svg,
#--var(tabs_id) .code a:hover > svg > * {
    stroke: --var(ui_color);
}

#--var(tabs_id) .code a:active > svg,
#--var(tabs_id) .code a:active > svg > * {
    stroke: #28a745;
}

#--var(tabs_id) p.info {
    color: #898989;
}

#--var(tabs_id) .instances-wrapper {
    max-height: 40rem;
}

#--var(tabs_id) .table-wrapper {
    overflow-y: scroll;
    background:
    linear-gradient(white 30%, rgba(255, 255, 255, 0)),
    linear-gradient(rgba(255, 255, 255, 0), white 70%) 0 100%,
    radial-gradient(50% 0, farthest-side, rgba(0, 0, 0, .2), rgba(0, 0, 0, 0)),
    radial-gradient(50% 100%, farthest-side, rgba(0, 0, 0, .2), rgba(0, 0, 0, 0)) 0 100%;
    background:
    linear-gradient(white 30%, rgba(255, 255, 255, 0)),
    linear-gradient(rgba(255, 255, 255, 0), white 70%) 0 100%,
    radial-gradient(farthest-side at 50% 0, rgba(0, 0, 0, .2), rgba(0, 0, 0, 0)),
    radial-gradient(farthest-side at 50% 100%, rgba(0, 0, 0, .2), rgba(0, 0, 0, 0)) 0 100%;
    background-repeat: no-repeat;
    background-color: white;
    background-size: 100% 40px, 100% 40px, 100% 14px, 100% 14px;
    background-attachment: local, local, scroll, scroll;
}

#--var(tabs_id) .table-wrapper table {
    table-layout: auto;
    width: 100%;
    font-size: 1em;
    border-collapse: collapse;
}

#--var(tabs_id) .table-wrapper tr:nth-child(odd) {
    background-color: rgba(0, 0, 0, 0.05);
}

#--var(tabs_id) .table-wrapper tr:hover {
    background-color: rgba(66, 165, 245, 0.2);
}

#--var(tabs_id) .table-wrapper td:nth-child(2),
#--var(tabs_id) .table-wrapper th:nth-child(2) {
    width: 99%;
}

#--var(tabs_id) .table-wrapper tr > td {
    text-align: left;
    padding: 1em;
    text-align: left;
}

#--var(tabs_id) .table-wrapper th {
    color: #fff;
    background-color: --var(ui_color);
}

#--var(tabs_id) .table-wrapper tr > th {
    text-align: left;
    padding: 1em;
    margin: -0.5em;
}

@media (min-width: 768px) {
    #--var(ui_id) {
        font-size: 1.33rem;
    }

    .ui-container {
        padding: 2rem 2rem;
    }

    #--var(tabs_id) label {
        order: 1;
        width: auto;
    }

    #--var(tabs_id) label.wide {
        flex: 1;
        align-items: left;
        justify-content: left;
    }

    #--var(tabs_id) .tab {
        order: 9;
    }

    #--var(tabs_id) [type=radio]:checked + label {
        border-bottom: none;
    }
}
"""
CUSTOM_JS = """
function copy(elem){
    var content = document.getElementById(elem).innerHTML;

    navigator.clipboard.writeText(content)
        .then(() => {
        console.log("Text copied to clipboard!")
    })
        .catch(err => {
        console.log('Something went wrong', err);
    })
}
"""


def format_label(label: str, label_name: str = 'Label', h: str = 'h3') -> str:
    """Format label as title.

    Args:
        label (str): Name of label
        label_name (str, optional): Label name. Defaults to 'Label'.
        h (str, optional): h-tag (h1, h2, ...). Defaults to 'h1'.

    Returns:
        str: Formatted label title.
    """
    return f'<{h}>{label_name.title()}: <kbd>{label}</kbd></{h}>'


def format_list(items: Iterable, format_fn: Optional[Union[Callable, str]] = None) -> str:
    """Format list of items to HTML.

    Args:
        items (Iterable): Items.
        format_fn (Optional[Union[Callable, str]], optional): How to format each element. If none just converts it to 
            a string. Defaults to None.

    Returns:
        str: Formatted list.
    """
    if format_fn is None:
        format_fn = lambda x: x  # noqa: E731
    elif isinstance(format_fn, str):
        _format_fn = copy.deepcopy(format_fn)
        format_fn = lambda x: f'<{_format_fn}>{x}</{_format_fn}>'  # noqa: E731
    return '<ul>' + ''.join(f'<li>{format_fn(str(item))}</li>' for item in items) + '</ul>'


def format_instance(instance: dict, **kwargs) -> str:
    """Format an `instancelib` instance.

    Args:
        instance (dict): `instancelib` instance exported to config.

    Returns:
        str: Formatted instance.
    """
    repr = instance['_representation'] if '_representation' in instance else instance['_data']
    identifier = instance['_identifier']
    instance_title = instance.get('__class__', '')
    optional_columns = ''.join(f'<td>{v}</td>' for v in kwargs.values())
    return f'<tr title="{instance_title}"><td>{identifier}</td><td>{repr}</td>{optional_columns}</tr>'


def format_instances(instances: Union[dict, List[dict]], **kwargs) -> str:
    """Format multiple `instancelib` instances.

    Args:
        instances (Union[dict, List[dict]]): instances.
        **kwargs: Optional named columns.

    Returns:
        str: Formatted instances.
    """
    if isinstance(instances, dict):
        instances = [instances]
    for k, v in kwargs.items():
        if isinstance(v, str):
            kwargs[k] = [v] * len(instances)
        elif isinstance(v, dict) and 'labelset' in v and 'labeldict' in v:
            kwargs[k] = {k_: ', '.join([f'<kbd>{v__}</kbd>' for v__ in v_]) for k_, v_ in v['labeldict'].items()}
        elif not isinstance(v, list):
            raise ValueError(f'Unable to parse {type(v)} ({v})')

    content = ''.join([format_instance(instance, **{k: v[i] if isinstance(v, list) else v[instance['_identifier']]
                                                    for k, v in kwargs.items()})
                       for i, instance in enumerate(instances)])
    header = ''.join([f'<th>{h}</th>' for h in ['ID', 'Instance'] + [str(k).title() for k in kwargs.keys()]])

    return f'<div class="instances-wrapper table-wrapper"><table><tr>{header}</tr>{content}</table></div>'


def is_colab() -> bool:
    """Check if the environment is Google Colab.

    Returns:
        bool: True if Google Colab, False if not.
    """
    return '.colab' in str(get_ipython().__class__).lower()


def is_interactive() -> bool:
    """Check if the environment is interactive (Jupyter Notebook).

    Returns:
        bool: True if interactive, False if not.
    """
    try:
        if 'interactive' in str.lower(get_ipython().__class__.__name__) or is_colab():
            return True
        return False
    except:  # noqa: E722
        return False


def internet_connection(url: str = 'http://www.pypi.org', timeout: int = 5) -> bool:
    """Check whether there is an active internet connection, by trying to reach an URL within timeout.

    Args:
        url (str, optional): URL to connect to. Defaults to 'http://www.pypi.org'.
        timeout (int, optional): Timeout. Defaults to 5.

    Returns:
        bool: True if has internet connection, else False.
    """
    import requests

    try:
        requests.get(url, timeout=timeout)
        return True
    except (requests.ConnectionError, requests.Timeout):
        return False


class Render:
    """Base class for rendering configs (configuration dictionaries)."""

    main_color = MAIN_COLOR
    package_link = PACKAGE_LINK
    extra_css = ''

    def __init__(self, *configs):
        """Initialize rendering class.

        Example:
            Writing your own custom rendering functions `format_title()` and `render_content()`, and give the tab 
            a custom title `default_title`, set the main UI color to red (`#ff0000`) and package link (URL in footer):

            >>> from genbase.ui.notebook import Render
            >>> class CustomRender(Render):
            ...     def __init__(self, *configs):
            ...         super().__init__(*configs)
            ...         self.default_title = 'My Custom Explanation'
            ...         self.main_color = '#ff00000'
            ...         self.package_link = 'https://git.io/text_explainability'
            ...
            ...     def format_title(self, title: str, h: str = 'h1', **renderargs) -> str:
            ...         return f'<{h} style="color: red;">{title}</{h}>
            ...
            ...     def render_content(self, meta: dict, content: dict, **renderargs):
            ...         type = meta['type'] if 'type' in meta else ''
            ...         return type.replace(' ').title() if 'explanation' in type else type
        """
        self.configs = self.__validate_configs(configs)
        self.default_title = 'Explanation'
        self.config_title = 'Config'
        self.main_color = Render.main_color
        self.package_link = Render.package_link
        self.extra_css = Render.extra_css

    def __validate_configs(self, *configs):
        configs = [li for subli in configs for li in subli]
        for config in configs:
            assert isinstance(config, dict), 'Config should be dict'  # nosec
            assert 'META' in config, 'Config should contain "META" key'  # nosec
            assert 'CONTENT' in config, 'Config should contain "CONTENT" key'  # nosec
        return configs

    @property
    def tab_title(self, **renderargs) -> str:
        """Title of content tab."""
        title = self.default_title
        titles = [config['META']['title'] for config in self.configs if 'title' in config['META']]
        if titles:
            title = ' | '.join(list(set(titles)))
        if 'title' in renderargs:
            title = renderargs['title']
        return title

    @property
    def custom_tab_title(self, **renderargs) -> str:
        """Title of custom tab."""
        return ''

    @property
    def package_name(self) -> str:
        """Name of package."""
        if hasattr(self, '_package_name'):
            return self._package_name
        return self.package_link.rstrip('/').split('/')[-1]

    @package_name.setter
    def package_name(self, package_name: str):
        self._package_name = package_name

    def css(self, **replacement_kwargs) -> str:
        """Dynamically fetch CSS.

        Returns:
            str: CSS.
        """
        css_ = CUSTOM_CSS + '\n' + self.extra_css
        for k, v in replacement_kwargs.items():
            css_ = css_.replace(f'--var({k})', v)
        return css_

    def format_title(self, title: str, h: str = 'h1', **renderargs) -> str:
        """Format title in HTML format.

        Args:
            title (str): Title contents.
            h (str, optional): h-tag (h1, h2, ...). Defaults to 'h1'.

        Returns:
            str: Formatted title.
        """
        return f'<{h}>{title}</{h}>'

    def format_subtitle(self, subtitle: str) -> str:
        """Format the subtitle in HTML format.

        Args:
            subtitle (str): Subtitle contents.

        Returns:
            str: Formatted subtitle.
        """
        return f'<p class="info">{subtitle}</p>'

    def render_title(self, meta: dict, content: dict, **renderargs) -> str:
        """Render the title as HTML. Overwrite this when subclassing for your custom implementation.

        Args:
            meta (dict): Meta config.
            content (dict): Content config.
            **renderags: Optional arguments for rendering.

        Returns:
            str: Formatted title.
        """
        title = renderargs.pop('title', None)
        if title is None:
            if 'title' in meta:
                title = meta['title']
            elif 'type' in meta:
                title = meta['type']
                if 'subtype' in meta:
                    title += f' ({meta["subtype"]})'

        return self.format_title(title, **renderargs) if title else ''

    def render_subtitle(self, meta: dict, content: dict, **renderargs) -> str:
        """Render a subtitle as HTML.

        Args:
            meta (dict): Meta information to decide on appropriate renderer.
            content (dict): Content to render.

        Returns:
            str: SUbtitle.
        """
        return self.format_subtitle(renderargs['subtitle']) if 'subtitle' in renderargs else ''

    def get_renderer(self, meta: dict):
        """Get a render function (Callable taking `meta`, `content` and `**renderargs` and returning a `str`).

        Args:
            meta (dict): Meta information to decide on appropriate renderer.
        """
        def default_renderer(meta, content, **renderargs):
            return f'<p>{meta}</p>' + f'<p>{content}</p>'
        return default_renderer

    def render_content(self, meta: dict, content: dict, **renderargs) -> str:
        """Render content as HTML. Overwrite this when subclassing for your custom implementation.

        Args:
            meta (dict): Meta config.
            content (dict): Content config.
            **renderags: Optional arguments for rendering.

        Returns:
            str: Formatted content.
        """
        renderer = self.get_renderer(meta)
        return renderer(meta, content, **renderargs)

    def render_elements(self, config: dict, **renderargs) -> str:
        """Render HTML title and content.

        Args:
            config (dict): Config meta & content.
            **renderags: Optional arguments for rendering.

        Returns:
            str: Formatted title and content.
        """
        meta, content = config['META'], config['CONTENT']
        return self.render_title(meta, content, **renderargs) + \
            self.render_subtitle(meta, content, **renderargs) + \
            self.render_content(meta, content, **renderargs)

    def custom_tab(self, config: dict, **renderargs) -> str:
        """Optionally render a custom tab."""
        return ''

    def as_html(self, **renderargs) -> str:
        """Get HTML element for interactive environments (e.g. Jupyter notebook).

        Args:
            **renderags: Optional arguments for rendering.

        Returns:
            str: HTML element.
        """
        def fmt_exception(e: Exception, fmt_type: str = 'JSON') -> str:
            res = f'ERROR IN PARSING {fmt_type}\n'
            res += '=' * len(res) + '\n'
            return res + '\n'.join(traceback.TracebackException.from_exception(e).format())

        try:
            json = '\n'.join(srsly.json_dumps(config, indent=2) for config in self.configs)
        except TypeError as e:
            json = fmt_exception(e, fmt_type='JSON')

        try:
            yaml = '\n'.join(srsly.yaml_dumps(config) for config in self.configs)
        except srsly.ruamel_yaml.representer.RepresenterError as e:
            yaml = fmt_exception(e, fmt_type='YAML')

        html = ''.join(self.render_elements(config, **renderargs) for config in self.configs)
        id = str(uuid.uuid4())
        tabs_id = f'tabs-{id}'
        ui_id = f'ui-{id}'

        CUSTOM_TAB = ''.join(self.custom_tab(config, **renderargs) for config in self.configs)
        if CUSTOM_TAB:
            CUSTOM_TAB = f"""<input type="radio" name="{tabs_id}" id="{tabs_id}-tab2"/>
                             <label class="wide" for="{tabs_id}-tab2">{self.custom_tab_title}</label>
                             <div class="tab">{CUSTOM_TAB}</div>"""

        HTML = f"""
        <div id="{ui_id}">
            <section class="ui-wrapper">
                <div class="ui-container">
                    <div class="ui-block">
                        <div id="{tabs_id}">
                            <input type="radio" name="{tabs_id}" id="{tabs_id}-tab1" checked="checked" />
                            <label class="wide" for="{tabs_id}-tab1">{self.tab_title}</label>
                            <div class="tab">{html}</div>
                            {CUSTOM_TAB}
                            <input type="radio" name="{tabs_id}" id="{tabs_id}-tab{'3' if CUSTOM_TAB else '2'}" />
                            <label for="{tabs_id}-tab{'3' if CUSTOM_TAB else '2'}">{self.config_title}</label>
                            <div class="tab code">
                                <section>
                                    <div class="pre-buttons">
                                <a class='copy' onclick="copy('json-output')" href="#" title="Copy JSON to clipboard">
                                    {CLONE_SVG}
                                </a>
                                    </div>
                                    <h3>JSON</h3>
                                </section>
                                <pre id="json-output">{json}</pre>

                                <section>
                                    <div class="pre-buttons">
                                <a class='copy' onclick="copy('yaml-output')" href="#" title="Copy YAML to clipboard">
                                    {CLONE_SVG}
                                </a>
                                    </div>
                                    <h3>YAML</h3>
                                </section>
                                <pre id="yaml-output">{yaml}</pre>
                            </div>
                        </div>
                    </div>
                </div>
            </section>
        </div>
        """

        JS = f'<script type="text/javascript">{CUSTOM_JS}</script>' if CUSTOM_JS else ''

        main_color = renderargs.pop('main_color', self.main_color)
        package = renderargs.pop('package_link', self.package_link)
        package_name = self.package_name
        add_plotly = renderargs.pop('add_plotly', False)

        PLOTLY = ''
        if add_plotly and 'plotly' in HTML:
            from plotly.offline import get_plotlyjs
            PLOTLY = f'<script type="text/javascript">{get_plotlyjs()}</script>'

        CSS = self.css(ui_color=main_color, ui_id=ui_id, tabs_id=tabs_id)
        FOOTER = f'<footer>Generated with <a href="{package}" target="_blank">{package_name}</a></footer>'

        if add_plotly:
            HTML = HTML.replace('require(["plotly"], function(Plotly) {', '').replace('});', '')

        return f'{PLOTLY}<style>{CSS}</style>{HTML}{FOOTER}{JS}'


if is_interactive() and plotly_available():
    from plotly.offline import init_notebook_mode
    init_notebook_mode(connected=internet_connection())



"""Mixins to enhance class functionality."""

from typing import Optional

import numpy as np


class SeedMixin:
    """Adds working with `._seed` and `._original_seed` for reproducibility."""

    @property
    def seed(self):
        """Seed for reproducibility."""
        return self._seed

    @seed.setter
    def seed(self, value: int):
        self._original_seed = value
        self._seed = value

    def reset_seed(self):
        """Reset the seed to the original seed value, and return self."""
        self._seed = self._original_seed
        return self

    def set_seed(self, seed: Optional[int] = None):
        """Set the current seed and original seed to a new value, and return self.

        Args:
            seed (Optional[int], optional): Seed value. If None, select a random seed. Defaults to None.
        """
        if seed is None:
            seed = np.random.randint(100000)
        self._original_seed = seed
        return self.reset_seed()


class CaseMixin:
    """Adds working with title-, sentence-, upper- and lowercase for random data generation."""

    def lower(self):
        """Switch to lowercase data generation, and return self."""
        self._lowercase = True
        self._sentencecase = False
        self._titlecase = False
        self._uppercase = False
        return self

    def sentence(self):
        """Switch to sentencecase data generation, and return self."""
        self._lowercase = False
        self._sentencecase = True
        self._titlecase = False
        self._uppercase = False
        return self

    def title(self):
        """Switch to titlecase data generation, and return self."""
        self._lowercase = False
        self._sentencecase = False
        self._titlecase = True
        self._uppercase = False
        return self

    def upper(self):
        """Switch to uppercase data generation, and return self."""
        self._lowercase = False
        self._sentencecase = False
        self._titlecase = False
        self._uppercase = True
        return self

    def original(self):
        """Switch to original case data generation, and return self."""
        self._lowercase = False
        self._sentencecase = False
        self._titlecase = False
        self._uppercase = False
        return self

    def apply_case(self, string):
        """Apply the selected case to a string."""
        if not isinstance(string, str) or string.isnumeric():
            return string

        if self._lowercase:
            return string.lower()
        elif self._sentencecase:
            return string.capitalize()
        elif self._titlecase:
            return string.title()
        elif self._uppercase:
            return string.upper()
        return string



"""Utility functions."""

import base64
import gc
import importlib.util
import pkgutil
import warnings
from importlib import import_module
from pathlib import Path
from types import ModuleType
from typing import Any, Dict, Iterator, List, Optional, Tuple
from zipfile import ZipExtFile

import numpy as np
import sklearn
import srsly
from instancelib.analysis.base import BinaryModelMetrics, MulticlassModelMetrics
from instancelib.environment.base import Environment
from instancelib.instances.base import Instance, InstanceProvider
from instancelib.labels.base import LabelProvider
from instancelib.machinelearning import AbstractClassifier

ModelMetrics = (BinaryModelMetrics, MulticlassModelMetrics)


def export_instancelib(obj) -> Dict[str, Any]:
    """`instancelib`-specific safe exports."""
    if isinstance(obj, Environment):
        return {'dataset': export_instancelib(obj.dataset),
                'labels': export_instancelib(obj.labels),
                'named_providers': export_instancelib(obj.named_providers)}
    elif isinstance(obj, Instance):
        return {k: export_safe(v) for k, v in obj.__dict__.items() if k not in ['_vector']}
    elif isinstance(obj, LabelProvider):
        return {'labelset': export_safe(obj._labelset),
                'labeldict': {k: export_safe(v) for k, v in obj._labeldict.items()}}
    elif isinstance(obj, InstanceProvider):
        return [export_instancelib(o) for o in obj.get_all()]
    elif isinstance(obj, ModelMetrics):
        return {k: export_safe(getattr(obj, k)) for k in dir(obj)
                if not k.startswith('_') and not callable(getattr(obj, k))}
    elif hasattr(obj, '__dict__'):
        return dict(recursive_to_dict(obj))
    return export_serializable(obj)


def export_serializable(obj):
    """Export in serializable format (`pickle.dumps()` that is `base64`-encoded).."""
    return base64.b64encode(srsly.pickle_dumps(obj))


def export_dict(nested: dict) -> Iterator[Tuple]:
    """Export a normal dictionary recursively.

    Args:
        nested (dict): Dictionary to export

    Yields:
        Iterator[Tuple]: Current level of key-value pairs.
    """
    for key, value in nested.items():
        if isinstance(value, dict):
            yield key, dict(export_dict(value))
        else:
            yield key, export_safe(value)


def export_safe(obj):
    """Safely export to transform into JSON or YAML."""
    if isinstance(obj, (int, np.integer)):
        return int(obj)
    elif isinstance(obj, (float, np.floating)):
        return float(obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif 'pandas' in str(type(obj)).lower() and hasattr(obj, 'to_dict') and callable(obj.to_dict):
        return obj.to_dict()
    elif isinstance(obj, str):
        return str(obj)
    elif isinstance(obj, (frozenset, set)):
        return [export_safe(o) for o in list(obj)]
    elif isinstance(obj, (list, tuple)):
        return [dict(recursive_to_dict(o)) if hasattr(o, '__dict__') else export_safe(o) for o in obj]
    elif isinstance(obj, dict):
        return dict(export_dict(obj))
    elif 'tensorflow.' in str(type(obj)).lower() or 'torch.' in str(type(obj)).lower():
        return 'TODO-EXPORT-TF-TORCH'
    elif callable(obj):
        return export_serializable(obj)
    return obj


def recursive_to_dict(nested: Any,
                      exclude: Optional[List[str]] = None,
                      include_class: bool = True) -> Iterator[Tuple[str, Any]]:
    """Recursively transform objects into a dictionary representation.

    Args:
        nested (Any): Current object.
        exclude (Optional[List[str]], optional): Keys to exclude. Defaults to None.
        include_class (bool, optional): Whether to include `__class__` (True) or not (False). Defaults to True.

    Yields:
        Iterator[Tuple[str, Any]]: Current level of key-value pairs.
    """
    exclude = [] if exclude is None else exclude
    if include_class and hasattr(nested, '__class__'):
        cls = str(nested.__class__).split("'")[1]
        if cls == 'type':
            yield '__name__', str(nested.__qualname__)
            return
        elif 'blackboxclassifier' in str.lower(cls):
            yield 'BLACKBOX', 'HAS_CONTENTS_HIDDEN_<3_SO_STOP_PEEKING_:)'
            return
        elif cls != 'dict':
            yield '__class__', cls
    if hasattr(nested, '__qualname__') and hasattr(nested, '__annotations__'):
        yield '__name__', str(nested.__qualname__)
        nested = nested.__annotations__
    elif hasattr(nested, 'to_config') and callable(nested.to_config):
        nested = nested.to_config()
    elif hasattr(nested, '__dict__'):
        nested = nested.__dict__
    if not hasattr(nested, 'items'):
        nested = dict(nested)
    for key, value in nested.items():
        if (isinstance(key, str) and not key.startswith('__')) and key not in exclude:
            if isinstance(value, (AbstractClassifier, Environment, Instance,
                                  InstanceProvider, LabelProvider, ModelMetrics)):
                yield export_safe(key), export_instancelib(value)
            elif isinstance(value, (sklearn.base.BaseEstimator)):
                yield export_safe(key), export_serializable(value)
            elif hasattr(value, '__dict__') or isinstance(value, dict):
                yield export_safe(key), dict(recursive_to_dict(value, exclude=exclude, include_class=include_class))
            else:
                yield export_safe(key), export_safe(value)


def get_file_type(pathlike: str) -> Optional[str]:
    """Get file type of a pathlike string.

    Args:
        pathlike (str): Pathlike string.

    Returns:
        Optional[str]: File extension or None.
    """
    if isinstance(pathlike, ZipExtFile):
        pathlike = pathlike.name
    if isinstance(pathlike, str):
        return str.lower(Path(pathlike).suffix)
    return None


def package_available(package: str) -> bool:
    """Check if package is installed.

    Args:
        package (str): Name of package.

    Returns:
        bool: Whether the package is available (True) or not (False).
    """
    return importlib.util.find_spec(package) is not None


def extract_metrics(metrics: dict) -> Tuple[dict, list]:
    """Extract metrics.

    Args:
        metrics (dict): Metrics.

    Returns:
        Tuple[dict, list]: Tuple containing extracted metrics and property names.
    """
    # Get all unique property names
    properties = []

    def safe_getattr(obj, name):
        try:
            return getattr(obj, name, None)
        except Exception as e:
            warning(e)
            return type(e).__name__

    for v in metrics.values():
        for p in dir(v):
            if not p.startswith('_') and not callable(safe_getattr(v, p)) and p not in properties:
                properties.append(p)

    # Extract property values
    def extract_property(m, p):
        res = safe_getattr(m, p)
        if isinstance(res, frozenset):
            res = len(res)
        return res

    return {k: {p: extract_property(v, p) for p in properties} for k, v in metrics.items()}, properties


def info(message):
    """Print an informational message."""
    warnings.formatwarning = lambda msg, *args, **kwargs: f'{msg}\n'
    warnings.warn(f'[INFO] {message}')


def warning(message):
    """Print a warning message."""
    warnings.formatwarning = lambda msg, *args, **kwargs: f'{msg}\n'
    warnings.warn(f'[EXCEPTION] {message}')


class silence_tqdm:
    """Silence all outputs of tqdm in a given module.

    Example:
        Silence tqdm in package `instancelib`:

        >>> import instancelib
        >>> with silence_tqdm(instancelib):
        ...    model.predict(instances)
    """

    def __init__(self, package: ModuleType):
        """
        Args:
            package (ModuleType): Module to silence tqdm in.
        """
        self.package = package
        self.orig_refs = {}

    def __enter__(self):
        """Override references to tqdm in `self.package` to just an iterator."""
        for _, name, _ in pkgutil.walk_packages(self.package.__path__, self.package.__name__ + '.'):
            try:
                referents = gc.get_referents(import_module(name))[0]
                if 'tqdm' in referents:
                    self.orig_refs[name] = referents['tqdm']
                    referents['tqdm'] = lambda i, *k, **kw: i
            except (AttributeError, ModuleNotFoundError, ValueError):
                pass

    def __exit__(self, *args):
        """Reset original references to tqdm in `self.package`."""
        for name, func in self.orig_refs.items():
            try:
                gc.get_referents(import_module(name))[0]['tqdm'] = func
            except (ModuleNotFoundError, ValueError):
                pass



"""Static & interactive plotting functions."""

import os
from typing import Callable

import pandas as pd
import plotly
import plotly.io as pio

from ...utils import package_available

ROOT_DIR = os.path.dirname(os.path.abspath(plotly.__file__))
PACKAGE_DIR = os.path.join(ROOT_DIR, 'package_data')
PLOTLYJS_FILE = os.path.join(PACKAGE_DIR, 'plotly.min.js')


def matplotlib_available() -> bool:
    """Check if `matplotlib` is installed.

    Returns:
        bool: True if available, False if not.
    """
    return package_available('matplotlib')


def plotly_available() -> bool:
    """Check if `plotly` is installed.

    Returns:
        bool: True if available, False if not.
    """
    import importlib.util
    return importlib.util.find_spec('plotly') is not None


class ExpressPlot:
    def __init__(self, df: pd.DataFrame, px_fn: Callable, *args, **kwargs):
        self.plot = px_fn(df, *args, **kwargs)

    @property
    def static(self):
        return self.to_png()

    @property
    def interactive(self) -> str:
        return self._repr_html_()

    def to_html(self, **kwargs) -> str:
        return pio.to_html(self.plot, **kwargs)

    def write_html(self, **kwargs):
        return pio.write_html(self.plot, **kwargs)

    def to_image(self, **kwargs):
        return pio.to_image(self.plot, **kwargs)

    def write_image(self, file: str, **kwargs):
        return pio.write_image(self.plot, file, **kwargs)

    def to_png(self, **kwargs):
        return pio.to_image(self.plot, format='png', **kwargs)

    def write_png(self, file: str, **kwargs):
        return pio.write_image(self.plot, file, **kwargs)

    def __str__(self) -> str:
        return str(self.plot)

    def _repr_html_(self) -> str:
        return pio.to_html(self.plot, full_html=False, include_plotlyjs='require')  # TODO: 'require' when offline



from distutils.util import convert_path
from os import path

import setuptools

main_ns = {}
with open(convert_path('genbase/_version.py')) as ver_file:
    exec(ver_file.read(), main_ns)  # nosec

with open(path.join(path.abspath(path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup( # type: ignore
    name = 'genbase',
    version = main_ns['__version__'],
    description = 'Generation base dependency',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    author = 'Marcel Robeer',
    author_email = 'm.j.robeer@uu.nl',
    license = 'GNU LGPL v3',
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
    url = 'https://git.science.uu.nl/m.j.robeer/genbase',
    packages = setuptools.find_packages(), # type : ignore
    include_package_data = True,
    install_requires = [
        'instancelib>=0.4.4.1',
        'instancelib-onnx>=0.1.3',
        'onnxruntime; python_version < "3.10"',
        'onnxruntime>=1.12.0; python_version >= "3.10"',
        'matplotlib>=3.5.0',
        'numpy>=1.19.5',
        'python-i18n>=0.3.9',
        'srsly>=2.4.2',
        'requests',
        'ipython>=7.30.1',
        'lazy-load>=0.8.2',
    ],
    python_requires = '>=3.8'
)



"""Data imports, sampling and generation."""

from typing import Callable, Dict, Iterator, List, Literal, Optional, Tuple, Union

import instancelib as il
import pandas as pd
from instancelib.typehints import KT, VT

from ..utils import get_file_type, info

Method = Literal['infer', 'glob', 'pandas']


METHODS = ['infer', 'glob', 'pandas']
PANDAS_FILE_TYPES = ['.csv', '.tsv', '.txt', '.json', '.pkl', '.xls', '.xlsx']



def get_compressed_files(ioargs):
    handle = ioargs.filepath_or_buffer
    mode = ioargs.mode
    compression = ioargs.compression.pop('method', None)

    if compression == 'gzip':
        import gzip

        if isinstance(handle, str):
            return gzip.GzipFile(filename=handle, mode=mode)
        return gzip.GzipFile(fileobj=handle, mode=mode)
    elif compression == 'bz2':
        import bz2
        return bz2.BZ2File(handle, mode=mode)
    elif compression == 'xz':
        from pandas.compat import get_lzma_file
        return get_lzma_file()(handle, mode=mode)
    elif compression == 'zip':
        from pandas.io.common import _BytesZipFile

        handle = _BytesZipFile(handle, mode)
        if handle.mode == "r":
            zip_names = handle.namelist()

            if len(zip_names) == 0:
                raise ValueError(f'Empty ZIP file "{ioargs.filepath_or_buffer}"')
            return [handle.open(name) for name in zip_names]
    raise NotImplementedError(f'Unable to process "{handle}" with compressiong method "{compression}"!')


def pandas_to_instancelib(dataset, data_cols, label_cols, label_map=None):
    env = il.pandas_to_env(dataset, data_cols, label_cols)
    if label_map is not None:
        if isinstance(label_map, dict):
            label_map = {str(k): v for k, v in label_map.items()}
        env = rename_labels(env, label_map)
    return env


def import_data(dataset,
                data_cols: Union[KT, List[KT]],
                label_cols: Union[KT, List[KT]],
                label_map: Optional[Union[Callable, dict]] = None,
                method: Method = 'infer',
                _to_instancelib: bool = True,
                **read_kwargs) -> Union[il.Environment, pd.DataFrame]:
    """Import data in an instancelib Environment.

    Examples:
        Import from an online .csv file with data in the 'text' column and labels in 'category':

        >>> from genbase import import_data
        >>> ds = import_data('https://storage.googleapis.com/dataset-uploader/bbc/bbc-text.csv',
                             data_cols='text', label_cols='category')

        Convert a pandas DataFrame to instancelib Environment:

        >>> from genbase import import_data
        >>> import pandas as pd
        >>> df = pd.read_csv('https://storage.googleapis.com/dataset-uploader/bbc/bbc-text.csv')
        >>> ds = import_data(df, data_cols='text', label_cols='category')

        Download a .zip file and convert each file in the zip to an instancelib Environment:

        >>> from genbase import import_data
        >>> ds = import_data('https://archive.ics.uci.edu/ml/machine-learning-databases/00462/drugsCom_raw.zip',
                             data_cols='review', label_cols='rating')

        Convert a huggingface dataset (sst2) to an instancelib Environment:

        >>> from genbase import import_data
        >>> from datasets import load_dataset
        >>> ds = import_data(load_dataset('glue', 'sst2'), data_cols='sentence', label_cols='label')

    Args:
        dataset (_type_): Dataset to import.
        data_cols (Union[KT, List[KT]]): Name of column(s) containing data.
        label_cols (Union[KT, List[KT]]): Name of column(s) containing labels.
        label_map (Optional[Union[Callable, dict]], optional): Label renaming dictionary/function. Defaults to None.
        method (Method, optional): Method used to import data. Choose from 'infer', 'glob', 'pandas'.
            Defaults to 'infer'.
        _to_instancelib (bool, optional): Whether to convert the final result to instancelib. Defaults to True.
        **read_kwargs: Optional arguments passed to reading call.

    Raises:
        ImportError: Unable to import file.
        ValueError: Invalid type of method.
        NotImplementedError: Import not yet implemented.

    Returns:
        Union[il.Environment, pd.DataFrame]: Environment for each file or dataset provided.
    """
    if method not in METHODS:
        raise ValueError(f'Unknown method "{method}", choose from {METHODS}.')

    if isinstance(data_cols, (int, str)):
        data_cols = [data_cols]
    if isinstance(label_cols, (int, str)):
        label_cols = [label_cols]

    file_type = get_file_type(dataset)
    path_like = isinstance(dataset, str)

    # Unpack archived file
    extensions = pd.io.common._compression_to_extension.values() if hasattr(pd.io.common, '_compression_to_extension') \
        else pd.io.common.extension_to_compression.keys()
    if file_type in extensions:
        ioargs = pd.io.common._get_filepath_or_buffer(dataset, compression=file_type.replace('.', ''))
        info(f'Unpacking file "{dataset}".')
        return import_from_key_values([(file.name, file) for file in get_compressed_files(ioargs)],
                                      data_cols=data_cols,
                                      label_cols=label_cols,
                                      label_map=label_map,
                                      method=method,
                                      **read_kwargs)

    # Infer method
    if method == 'infer':
        if path_like and '*' in dataset:
            method = 'glob'
        elif file_type in PANDAS_FILE_TYPES:
            method = 'pandas'

    # Multiple files
    if method == 'glob':
        import glob
        return import_from_key_values([(file, file) for file in glob.glob(dataset)],
                                      data_cols=data_cols,
                                      label_cols=label_cols,
                                      label_map=label_map,
                                      method='infer',
                                      **read_kwargs)

    # Read one file with Pandas
    if method == 'pandas':
        if file_type is not None:
            info(f'Reading file "{dataset}".')
            if file_type in ['.csv', '.tsv', '.txt']:
                if 'sep' not in read_kwargs:
                    if file_type == '.csv':
                        read_kwargs['sep'] = ','
                    elif file_type == '.tsv':
                        read_kwargs['sep'] = '\t'
                dataset = pd.read_csv(dataset, **read_kwargs)
            elif file_type == '.json':
                dataset = pd.read_json(dataset, **read_kwargs)
            elif file_type == '.pkl':
                dataset = pd.read_pickle(dataset, **read_kwargs)
            elif file_type in ['.xls', '.xlsx']:
                dataset = pd.read_excel(dataset, **read_kwargs)
            else:
                raise ImportError(f'Unable to process file type "{file_type}" with method "pandas"!')

    if hasattr(dataset, 'to_pandas') and callable(dataset.to_pandas):
        info(f'Preparing "{dataset}" for import with Pandas.'.replace('\n', ' ').replace('\t', ''))
        dataset = dataset.to_pandas()
    elif isinstance(dataset, dict):
        return import_from_key_values(dataset.items(),
                                      data_cols=data_cols,
                                      label_cols=label_cols,
                                      label_map=label_map,
                                      **read_kwargs)

    if _to_instancelib:
        return pandas_to_instancelib(dataset, data_cols=data_cols, label_cols=label_cols, label_map=label_map)
    return dataset


def import_from_key_values(iterator: Iterator[Tuple[KT, VT]],
                           data_cols: Union[KT, List[KT]],
                           label_cols: Union[KT, List[KT]],
                           label_map: Optional[Union[Callable, dict]] = None,
                           method: Method = 'infer',
                           **read_kwargs) -> Dict[KT, il.Environment]:
    dataset = {k: import_data(v, data_cols=data_cols, label_cols=label_cols,
                              method=method, _to_instancelib=False, **read_kwargs)
               for k, v in iterator}
    return pandas_to_instancelib(dataset, data_cols=data_cols, label_cols=label_cols, label_map=label_map)


def train_test_split(environment: il.Environment,
                     train_size: Union[int, float],
                     train_name: str = 'train',
                     test_name: str = 'test') -> il.Environment:
    """Split an environment into training and test data, and save it to the original environment.

    Args:
        environment (instancelib.Environment): Environment containing all data (`environment.dataset`), 
            including labels (`environment.labels`).
        train_size (Union[int, float]): Size of training data, as a proportion [0, 1] or number of instances > 1.
        train_name (str, optional): Name of train split. Defaults to 'train'.
        test_name (str, optional): Name of train split. Defaults to 'test'.

    Returns:
        instancelib.Environment: Environment with named splits `train_name` (containing training data) and `test_name`
            (containing test data) 
    """
    environment[train_name], environment[test_name] = environment.train_test_split(environment.dataset,
                                                                                   train_size=train_size)
    return environment


def rename_labels(provider: Union[il.Environment, il.LabelProvider],
                  mapping: Union[Callable, dict]) -> Union[il.Environment, il.LabelProvider]:
    """Rename labels in a labelprovider or environment.

    Args:
        provider (Union[il.Environment, il.LabelProvider]): Provider to rename labels in.
        mapping (Union[Callable, dict]): Rename function or dictionary containing label mapping.

    Returns:
        Union[il.Environment, il.LabelProvider]: Original provider with labels remapped.
    """
    is_environment = isinstance(provider, il.Environment)
    _provider = provider.labels if is_environment else provider
    _provider = il.MemoryLabelProvider.rename_labels(_provider, mapping)
    if is_environment:
        provider._labelprovider = _provider
    else:
        provider = _provider
    return provider
